
sub fixLocales {
	foreach my $e (keys %LocaleEnvs){
		if (exists $ENV{$e}) {
			$LocaleEnvs{$e} = $ENV{$e};
			$ENV{$e} = "en_US.UTF-8";
			#delete $ENV{$e};
		}
	}
}

sub resetLocales {
	foreach my $e (keys %LocaleEnvs){
		if (defined $LocaleEnvs{$e}){
			$ENV{$e} = $LocaleEnvs{$e};
		}
	}
}

sub wLog {
	return if ($WARNABLE > 0);
	my (@warn) = @_;
	chomp $warn[$#warn];
#	push @warn, " (ignore: $WARNABLE)" if ($WARNABLE > 0);
	print STDERR "\e[1mW: ",@warn,"\e[0m\n"; 
#	ECHO("$APTFU/aptfu.wLog", @warn);
	push @WARNINGS, join('', @warn);
}

sub ECHO {
	my $f = shift;
	open EE, ">> $f" || EXIT (">> $f (echo @_)");
	print EE @_,"\n";
	close EE;
}

sub iLog {
	my (@imsg) = @_;
	chomp $imsg[$#imsg];
	print 'I: ',@imsg,"\n"; 
	push @INFOMSGS, join('', @imsg);
#	ECHO("$APTFU/aptfu.iLog", @imsg);
}

sub checkSymlink {
	my $s = shift;
	my $reals = readlink($s);
	if (!(defined $reals) || !(-r $reals)) {
		warn "DEAD symlink: $s REMOVED\n";
		rx("unlink $s");
		return 1;
	}
	return 0;
}

sub versionList {
	my $pkg = shift;
	my @ret;
	my %in;
	show2Cache($pkg);
	foreach my $type (split(/\s*\,\s*/,$ENV{'DEPORDER'})) {
		my $v = $verDefault;
		if ($type eq 'current'){
			$v = currentVer($pkg);
		} elsif ($type eq 'local'){
			$v = localVer($pkg);
		} elsif ($type eq 'default'){
			$v = P_verDefault($pkg);
		} elsif ($type eq 'newest'){
			$v = P_verNewest($pkg);
		}
		next if ($in{$v} || ($v eq $verDefault));
		push @ret, $v;
		$in{$v} = 1;
	}
	return @ret;
}

sub checkDepVer {
	my ($depPkg,$depExpr,$depVer) = @_;
	
	show2Cache($depPkg);

	foreach my $v ( versionList($depPkg),
	sort{ $VS->compare($b,$a)}(keys %{$PCACHE->{$depPkg}}) ){
		next if (defined $CacheField{$v});
#		print "VVV $v $depExpr $depVer\n" if ($depPkg =~/libnet-perl/);
		if (($v ne $verDefault) 
		&& $VS->check_dep($v,$depExpr,$depVer)) {
			return $v;
		}
#		print "FVV $v $depExpr $depVer\n" if ($depPkg =~/libnet-perl/);
	}
	return $verDefault;
}


#sub BDinfo {
#	my ($p,$v) = @_;
#	($p,$v) = S_verSelect($p,$v);
#	my $np= namePrefix($p);
#	chdir $CACHEDIR or EXIT("$! chdir $CACHEDIR");
#	foreach my $f (<*>) {
#		next unless ($f=~/^build-info/);
#		my $fh = zOpen($f);
#		while (<$fh>) {
#			
#		}
#		close $fh;
#	}
#}
#
# returns the first character of a package name or the first 4 if it's ^lib\w
sub namePrefix {
	my $prefix = shift;
	if ($prefix =~ /^(lib.)/) { $prefix = $1 }
	else { $prefix =~ s/^([a-z0-9]).*$/$1/ }
	return $prefix;
}

sub mchomp { foreach (@_) { next unless ($_); s/^\s+//g; s/\s+$//g; } }

sub lvchomp { foreach (@_) { next unless ($_); s/\.aptfu\.\d+\s*$//; } }

sub readChangelog{
	if (-r $_[0]) {
		my $fh = zOpen($_[0]);
		while (<$fh>){
			if (/^(\S+)\s+\((\S+)\)/){
				close $fh;
				return ($1,$2);
				last;
			}
		}
		close $fh;
	}
	EXIT("$_[0] unreadable/unparseable in ", cwd);
}

sub md5sum {
	my $fn = shift;
	my $md5o = Digest::MD5->new || EXIT("Can't initialize MD5 ($fn)");
	open MD, $fn || EXIT("Failed to open $fn");
	$md5o->addfile(*MD);
	my $ret = $md5o->hexdigest;
	close MD;
	return $ret;
}

sub info { 
	my ($lvl,@msg) = @_;
	$msg[$#msg] .= "\n" unless ($msg[$#msg] =~/\n$/);
	print @msg if ( $lvl <= $opts{'verbose'} );
}

sub infon { 
	my ($lvl,@msg) = @_;
	print @msg if ( $lvl <= $opts{'verbose'} );
}

sub dbg { 
	my ($m,@dbg) = @_;
	if ($opts{'dbg'} & $m){
		$dbg[$#dbg] .= "\n" unless ($dbg[$#dbg] =~/\n$/);
		printf "[0x%08x] @dbg",$m;
	}
}

sub sys { info(2,"@_ $RDR"); return system("@_ $RDR"); }

sub rx { info(2,"$ROOTX @_"); return system("$ROOTX @_"); }

sub myDate { return strftime("%Y%m%d-%H%M%S",(localtime($_[0]))); }

sub compress {
	return 0 if ($DZ =~/(?:\W|^)cat$/);
	sys("$DZ --force @_") == 0 or EXIT("$! $DZ @_");
	return 0;
}

# use the control file to parse build-dependency information (method preferred)
# always favor this over apt-cache showsrc or AptPkg::Source
sub aptGetBD_ctrl {
	my ($ctrl) = @_;
	my ($fh);
	my $ret = {
		'Source' => '',
		'Build-Depends' => '',
		'Build-Depends-Indep' => '',
		'Build-Conflicts' => '',
		'Build-Conflicts-Indep' => '',
	},
	$fh = zOpen($ctrl)||EXIT("$! error opening $ctrl\n");
	$ret = ctrl2Href($fh,$ret,$ret,'control');
	# should be only one $s and $v to iterate through
#	foreach (keys %$ret) {	print "POO $_ $ret->{$_}\n";}

	return ($ret->{'Build-Depends'}, $ret->{'Build-Conflicts'},
	$ret->{'Build-Depends-Indep'}, $ret->{'Build-Conflicts-Indep'});
}

sub splitVer {
	lvchomp($_[0]);
	my ($ep,$ups,$deb) = ($_[0] =~ /^([^:]*:)?(.*?)(-[^-]*)?$/);
	$ep = '' unless (defined $ep);
	$deb = '' unless (defined $deb);
	return ($ep,$ups,$deb);
}

sub localListName {
	my $apt_list = $CACHEDIR;
	$apt_list =~ tr#/#_#;
	$apt_list .= "_._Packages";
	$apt_list = "$APT_LISTDIR/$apt_list";
	rx("touch $apt_list") unless (-f $apt_list);
	return $apt_list;
}

# automatic detection and open routine for compressed files
sub zOpen {
	my $F = shift;
	my $fh;
	EXIT("$F not found or readable") unless (-r $F);
	my ($cat,$p) = zCat($F);
	if ($cat && $p) {
		open $fh, "$cat $F $p" ||EXIT("$!: unable to ($cat $p) $F");
	} else {
		open $fh, $F ||EXIT("$!: unable to (zOpen) open $F");
	}
	return $fh;
}

sub zPatch {
	my ($F,$l,$R) = @_;
	$l = defined $l ? $l : 0;
	$R = defined $R ? $R : '';
	my $limit = $l + 10;
	my ($cat,$p) = zCat($F);
	print "Trying patch: $F, level ";
	my $i = 0;
	for ($i = $l; $i <= $limit; $i++){
		my $ret;
		print $i;
		if ($cat && $p) {
			last if system("$cat $F | patch $R --forward -p$i --force --dry-run >/dev/null 2>&1") == 0;
		} else {
			last if system("patch $R --force --forward --dry-run -p$i < $F >/dev/null 2>&1") == 0;
		}
	}
	print "\n";
	return if ($opts{'simulate'});
	print "Patching at $i\n";
	if ($cat && $p) {
		system("$cat $F | patch $R --forward -p$i") == 0
		||EXIT("$! $cat $F | patch $R --forward -p$i");
		return $i;
	} else {
		system("patch $R --forward -p$i < $F") == 0
		||EXIT("$! patch $R --forward -p$i < $F");
		return $i;
	}
}

sub zView {
	my ($F) = @_;
	my ($cat,$p) = zCat($F);
	if ($cat && $p) {
		return system("$cat $F $p sensible-pager");
	}
	return system("sensible-pager $F");
}

sub zCat {
	if ($_[0] =~ /(?:\.gz)|(?:\-gz)|(?:\-z)|(?:\.Z)|(?:\.z)|(?:_z)$/){
		return ('zcat','|');
	} elsif ($_[0] =~/\.bz2$/) {
		return ('bzcat','|');
	} else {
		return (undef,undef);
	}
}
	
1;

