sub swapVirtualPkg {
	my $pkg = shift;
	if ((isInstalled($pkg) != 0) && (defined $ALT{$pkg})) { 
		$pkg = $ALT{$pkg};
	}
	return $pkg;
}	

sub resolvePrint {
	my ($str,$V) = @_;
	$str =~ s/^\s*[A-Z\-]+\:\s*//i;
	foreach my $d (split(/\s*\,\s*/,$str)){
		print "$d => " if ($opts{'verbose'}>0);
		if (my $dep = resolveDep($d,$V)){
			print "$dep\n";
			push (@ret,$dep);
		} elsif ($opts{'verbose'}>0){
			print "undefined for $XARCH{$ENV{'MY_DEBIAN_ARCH'}}/$ENV{'MY_DEBIAN_ARCH'}\n";
		}
	}
	return @ret;
}

sub resolveVP {
	my ($pkg,$V) = @_;
	my $prov = getVirtuals($pkg);
	if ($prov) {
		my $default = '';
		my $count = 0;
		foreach my $i (sort {$a <=> $b} keys %$prov) {
			#dbg(0x1000000, "$prov->{$i}->{'P'}\n");
			if (isInstalled($prov->{$i}->{'P'}) == 0) {
				my $ver;
				$ver = currentVer($prov->{$i}->{'P'});

				if ($V) {
					$default .= $prov->{$i}->{'P'}
						.'='.$ver.' ';
				} else {
					$default = $prov->{$i}->{'P'}
						.'='.$ver;
					last;
				}
			}
			++$count;
		}
		unless ($default) {
			foreach my $fu ( sort  {$a <=> $b} keys %$prov ) {
				# grab the first one that resolves
				# and use it as default
				if (my $k = resolveDep($prov->{$fu}->{'P'},$V)){
					$default = $k;
					last;
				}
			}
		}
		if ($prompt{'virtual'} && ($count > 1) && (!$V) ) {
			my $dnum = '';
			my $sel;
			print "The following packages provide $pkg.  ",
				"Select one\:\n";
			foreach my $i (sort  {$a <=> $b} keys %$prov) {
				printf "\t%2i. %s (%s)\n",$i,
					$prov->{$i}->{'P'},$prov->{$i}->{'V'};
				$sel->{$i}->{'P'} = $prov->{$i}->{'P'};
				$sel->{$i}->{'V'} = $prov->{$i}->{'V'};
				my $tmp = $prov->{$i}->{'P'}
					.'='.$prov->{$i}->{'V'};
				if ( $tmp eq $default ){
					$dnum="$i. ".$prov->{$i}->{'P'}.' ('.
						$prov->{$i}->{'V'}.')';
				}
			}
			$dnum = $dnum ? $dnum : $default;
			print "Select one. [default=$dnum]: ";
			mchomp ( my $pick = <STDIN> );
			foreach ( split /\D+/,$pick ){
				if ( defined $sel->{$_}->{'P'} 
				&& ( $sel->{$_}->{'P'} 
				eq resolveDep($sel->{$_}->{'P'},$V) ) ){
					return $sel->{$_}->{'P'}
						.'='.$sel->{$_}->{'V'};
				}
			}
		}
		return $default;
	}
#	unless ($V) {
#		wLog("Virtual packages for '$pkg' failed to resolve!");
#	}
	return undef; 
}

# resolveDep() - I made this work since apt-get build-dep $pkg 
# doesn't always work for |, [arch], and virtual packages
# return values for resolveDep()
# $V is for inVert, used for Conflicting packages
# returns a $string with package=version (can be multiple packages=versions if 
# inverted for conflicts

sub resolveDep {
	my ($dep,$V) = @_;
	
	unless ($dep =~/\S/) { 
		warn "No Package specified for Dependency resolution!";
		return undef;
	}
	
	# give everything a consistent format to make caching easier...
	mchomp($dep);
	$dep =~ s/(\||\)|\])(\S)/$1 $2/g;
	$dep =~ s/(\S)(\||\(|\[)/$1 $2/g;
	$dep =~ s/\s+/ /g;
	$dep =~ s/\(\s*([\=\>\<]+)\s*([\.\d\+\-\w\:]+)\s*\)/\($1 $2\)/g;
	if ($dep =~/\[([^\]]+)\]/) {
		my $old = $1;
		mchomp(my $new = $old);
		$old = join(' ',sort(split(/\s+/,$new)));
		$dep =~s/\[\Q$old\E\]/\[$new\]/;
	}
	
	if ($DEPCACHE->{"$dep $V"} && ($DEPCACHE->{"$dep $V"} ne 'undef')) {
		return $DEPCACHE->{"$dep $V"};
	} elsif ($DEPCACHE->{"$dep $V"}){
		return undef;
	}
	 
	if ($dep =~/\|/) { 
	#  one|two|three... pick one of the pickable options, try to choose
	#  one that's already installed, first.
		my @pipes = split /\|/, $dep;
		$dep = join(' | ', sort(@pipes));
		++$WARNABLE;
		my @np;
		foreach my $p (@pipes) {
			if (my $res = resolveDep($p,$V) ) {
				#dbg(0x200,"$dep -> $res\n");
				if (isVInstalled(split(/=/,$res)) == 0) {
					--$WARNABLE;
					$DEPCACHE->{"$dep $V"} = $res;
					return $res;
				}
				push @np,$res;
			}
		}
		--$WARNABLE;
		foreach my $nn (@np) {
			if ($GTEMP->{$nn}){
				$DEPCACHE->{"$dep $V"} = $nn;
				return $nn;
			}
		}
		foreach my $nn (@np) {
			if (localAvail(split(/=/,$nn)) == 0){
				$DEPCACHE->{"$dep $V"} = $nn;
				$GTEMP->{$nn} = 1;
				return $nn;
			}
		}
		return $np[0] if ($np[0]);
		unless ($V) {
			wLog("None of these (@pipes) dependencies resolved!");
		}
		#dbg(0x200,"Failed Resolving package $dep\n");
		$DEPCACHE->{"$dep $V"} = 'undef';
		return undef;
	} 
	if ($dep =~/\[([^\]]+)\]/) {
		my @tmps = split /\s+/, $1;
		my $not = 0;
		foreach my $i (@tmps) {
			if ($i =~ /\!/ && $not == 0) {
				$not = 1;
			}
			if ($i eq $XARCH{$ENV{'MY_DEBIAN_ARCH'}} 
			|| $i eq $ENV{'MY_DEBIAN_ARCH'}) { 
				last;
			} elsif ($i eq '!'.$XARCH{$ENV{'MY_DEBIAN_ARCH'}} 
			|| $i eq '!'.$ENV{'MY_DEBIAN_ARCH'}) {
				#dbg(0x200,"$dep ignored for our ARCH!\n");
				$DEPCACHE->{"$dep $V"} = 'undef';
				return undef;
			} elsif ( $i eq $tmps[$#tmps]){
				last if ($not);
				#dbg(0x200,"$dep ignored for our ARCH!\n");
				$DEPCACHE->{"$dep $V"} = 'undef';
				return undef;
			}
		}
		#fall through if it made it past the arch section
	}

	my ($pkg) = ($dep =~ /^([\d\w\-\.\+]+)/);
	# ok, we got past the architecture-specific stuff and | stuff
	# try to resolve virtual package here if user has 
	# specified a replacement
	unless ($V) {
		my $vp = swapVirtualPkg($pkg);
		$dep = $vp if ($pkg ne $vp);
	}
	my $retver = $verDefault;
	foreach my $va (versionList($pkg)){
		$retver = $va if ($retver eq $verDefault);
		last if ($retver ne $verDefault);
	}
	if ($retver eq $verDefault){
	#dbg(0x200,"Virtual packages specified, need to pick one\n");
		my $tret = resolveVP($pkg,$V);
		$DEPCACHE->{"$dep $V"} = $tret;
		return $tret;
	}
	if ($dep =~/\(/) { 
	# we have version information to parse
	# "binutils (>= 2.13.90.0.10) [!m68k !sparc]" would become
	# "binutils >= 2.13.90.0.10"
	# arch information has already been processed above so it's stripped
	# 1: PackageName  2: cmpExpression(optional) 3: VersionRequirement
		my ($depExpr,$depVer) = ($dep =~ /^(?:[\d\w\-\.\+]+)\s*
			\(\s*([\=\<\>]*?)\s*([\.\d\+\-\w\:]+?)\)/x );
		unless ($depExpr) { $depExpr = '=' }
		unless ($depVer) { wLog("DEP: $dep has no version info!") }
		$retver = checkDepVer($pkg,$depExpr,$depVer);
		# include versions in resolveDep
		if (($retver eq $verDefault) && !$V) {
			wLog("no available version satisfies $dep [$pkg $depExpr $depVer]");
			$DEPCACHE->{"$dep $V"} = 'undef';
			return undef;
		}

	}
	mchomp ($pkg);
	$DEPCACHE->{"$dep $V"} = "$pkg=$retver";
	return "$pkg=$retver";
} #endsub resolveDep()


sub rmLoop {
	my ($tmp,$rms,$deps);
	%$rms = map {$_ => 1} @_;
	unless (defined $opts{'min-prio'} && defined $PRIO{$opts{'min-prio'}}){
		$opts{'min-prio'} = 'standard';
	}

	info(0,"Calculating recursive removal of\n  @_\n",
	"and associated dependencies. ",
	"This may take a while.\n");

	# get the reverse dependencies of the specified
	# packages.  These will automatically be cleared since the user
	# indirectly specified their removal
	#
	# in short, stuff that goes into $tmp in the next line 
	# are definitely going to get passed to `apt-get remove`
	foreach my $r (keys %$rms) { 
		$tmp->{$r} = getRevDeps($r);
	}
	foreach my $t (keys %$tmp) { 
		foreach my $i (keys %{$tmp->{$t}}){
			$rms->{$i} = 1;
			#dbg(0x10000,"ADDED: $_\n");
			$deps->{$i} = 1;
		}
	}
	
	my $last = '';
	my $now = join ' ', sort keys %$rms;

	while ( $now ne $last ) {
		# XXX this is slow, and not thorough enough
		#dbg(0x10000,"$APTGET remove $now\n");
		my $last2 = '';
		my $now2 = join ' ', sort keys %$deps;
		while ( $now2 ne $last2 ) {
			# get install dependencies of all packages
			my $rm = getInDeps($deps);

			# resolve app the dependencies gotten.
			$deps = resolveRevDeps($deps,$rm);
			$last2 = $now2;
			$now2 = join ' ', sort keys %$deps;
		}

		$rms = depFilter($deps,$rms);
		$last = $now;
		$now = join ' ',sort keys %$rms;
		#dbg(0x10000,"$APTGET remove $now\n");
	}
	return $now;
}


sub resolveRevDeps {
	my ($deps,$rm) = @_;
	foreach my $p (keys %$rm) {
		foreach my $d (keys %{$rm->{$p}}) {
			my $v; # version probably won't be useful
			($d,$v) = split /=/, resolveDep($d,0);
			$deps->{$d} = 1 if ($d);
		}
	}
	return $deps;
}


sub resolveInDeps{
	my ($deps,$rm) = @_;
	foreach my $p (keys %$rm) {
		foreach my $d (keys %{$rm->{$p}}) {
#			print "XXXXXXXXX ResolveInDeps $d";
			if (my $v = resolveDep($d,0)){
				$deps->{$v} = 0 unless (defined $deps->{$v});
				# priority calculation
				if ($deps->{$v} <= 0){
					--$deps->{$v};
				} else {
					++$deps->{$v};
				}
			}
		}
	}
	return $deps;
}

sub getInDeps {
	my $rm = ();
	foreach my $pkg (keys %{$_[0]}) {
		if ($_[0]->{$pkg} > 0){
			# priority calculation
			++$_[0]->{$pkg};
			next;
		}
		if ($pkg =~ /([\d\w\-\.\+]+)\=([\d\w\-\.\+\:]+)/) {
			$rm->{$pkg} = aptGetID($1,$2);
		} elsif (isInstalled($pkg) == 0){
			$rm->{$pkg} = aptGetID($pkg,currentVer($pkg));
		} else {
			$rm->{$pkg} = aptGetID($pkg,P_verDefault($pkg));
		}
		$_[0]->{$pkg} = abs($_[0]->{$pkg});
		++$_[0]->{$pkg};
	}
	return $rm;
}

sub depFilter {
	my ($deps,$rms) = @_;
	foreach my $d (sort keys %$deps) {
		# move along if we have a redeclared pkg
		next if ($rms->{$d});
		
		if (my $tmps = getRevDeps($d)){
			my $i = 0;
			foreach my $t (sort keys %$tmps) {
				if (P_getPrio($t,currentVer($t))
				>= $PRIO{$opts{'min-prio'}}){
					#dbg(0x20000,"TIPPED: $d $t\n");
					++$i;
					last;
				}
				if (($t eq $d) || $rms->{$t}) {
					#dbg(0x20000,"IGNORED: $d $t\n");
					next;
				}
				if ($deps->{$t}){
					if (revDepFilter($t,$deps)){
						#dbg(0x20000,"RIPPED: $d $t\n");
						++$i;
						last;
					} else {
						#dbg(0x20000,"DGNORED: $d $t\n");
						next;
					}
				}
				#dbg(0x20000,"TIPPED: $d $t\n");
				++$i;
				last;
			}
			# the ones without reverse depends get removed
			if ($i == 0) {
				#dbg(0x20000,"ADDED: $d\n");
				$rms->{$d} = 1;
			}
		} elsif (P_getPrio($d,currentVer($d)) < $PRIO{$opts{'min-prio'}}) {
			#dbg(0x20000,"ADDED: $d\n");
			$rms->{$d} = 1;
		}
	}
	return $rms;
}

sub revDepFilter {
	my ($pkg,$filter) = @_;
	my $ret = ();
	if (my $tmp = getRevDeps($pkg)) {
		return 1;
	}
	return 0;
}

sub revDepTree {
	my $pee = shift;
	my $poo;
	foreach my $p (keys %$pee) {
		mchomp($p);
		if ($p =~ /^[\d\w\-\.\+]+\=[\d\w\-\.\+\:]+$/) {
			if (isVInstalled(split(/=/,$be))==0) {
				$poo->{$p} = 0;
				next
			}
		} elsif (my $be = resolveDep($p,1) ){
			if ($be &&(isVInstalled(split(/=/,$be))==0)) {
				$poo->{$be} = 0;
			}
		}
	}
	return unless (keys %$poo);
	$opts{'verbose'} -= 2;
	my ($inp,$rmp,$dirty) = simAptGet('remove',join(' ',keys(%$poo)),0);
	$opts{'verbose'} += 2;
	if ($inp) {
		wLog("Package(s) \n  ", join(' ',keys(%$inp)),
		"\nwill be installed if \n  ",
		join(' ',keys(%$rmp)),"\nis/are installed");
	}
	return ($rmp);
}

sub TopLvlDeps {
	my $pee = shift;
	my $poo;
	foreach my $p (keys %$pee) {
		mchomp($p);
		if ($p =~ /^[\d\w\-\.\+]+\=[\d\w\-\.\+\:]+$/) {
			$poo->{$p} = 0;
		} else {
			if (my $be = resolveDep($p,0)){
				$poo->{$be} = 0;
			}
		}
	}
	policy2CacheM($in);
	my $in = getInDeps($poo);
	$poo = resolveInDeps($poo,$in);
	policy2CacheM($poo);
#	foreach (keys %$poo) {print "$_\n";}	print "\n\n";
	return $poo;
}

# get all installed dependencies
sub testDepTree {
	my $pee = shift;
	my $poo;
	foreach my $p (keys %$pee) {
		mchomp($p);
		if ($p =~ /^[\d\w\-\.\+]+\=[\d\w\-\.\+\:]+$/) {
			$poo->{$p} = 0;
		} else {
			if (my $be = resolveDep($p,0)){
				$poo->{$be} = 0 if ($be);
			}
		}
	}
	my $last = '';
	my $now = 'Q';
	while ($now ne $last) {
		policy2CacheM($poo);
		my $in = getInDeps($poo);
		policy2CacheM($in);
		$GTEMP = $poo;
		$poo = resolveInDeps($poo,$in);
		$last = $now;
		$now = join '', sort keys %$poo;
	}
	undef $GTEMP;
	return $poo;
}

1;

