

# only call this on a real source package name, not a binary package name
# ( there's no need to on a binary name)
sub getFuVersion {
	my ($p,$v) = @_;
	my $rev = 1;
	if ($v=~s/\.aptfu\.(\d+)$//) {
		$rev += $1;
#		info(0,"GFV: $v.aptfu.$rev");
		return "$v.aptfu.$rev";
	}
	showsrc2Cache($p);
	return "$v.aptfu.$rev" unless (defined $SCACHE->{$p});
	if ((defined $SCACHE->{$p}->{$v}) 
	&& (my $bin = $SCACHE->{$p}->{$v}->{'Binary'})){
		foreach my $pp (split(/\s*(?:,|\n)\s*/,$bin)){
			chomp $pp;
			my $cv = currentVer($pp);
			my $lv = localVer($pp);
			
			my @vs = ( $cv, $lv, binSourceVField($pp,$cv),
			binSourceVField($pp,$lv));
			
			foreach (@vs) {
				if (s/\.aptfu\.(\d+)$//) {
					if (($_ eq $v) && ($1 >= $rev)){
						$rev = $1 + 1;
					} 
				}
			}
		}
	}
	return "$v.aptfu.$rev";
}

sub aptGetBD_cache {
	my ($src,$ver) = @_;
	my ($bd,$bc,$bdi,$bci) = ('','','','');
	lvchomp($ver);
	# $src is already a source package name, not a package->bin remap
	showsrc2Cache($src);
	if (defined $SCACHE->{$src}->{$ver}->{'Build-Depends'}) {
		$bd = $SCACHE->{$src}->{$ver}->{'Build-Depends'};
	}
	if (defined $SCACHE->{$src}->{$ver}->{'Build-Conflicts'}) {
		$bc = $SCACHE->{$src}->{$ver}->{'Build-Conflicts'};
	}
	if (defined $SCACHE->{$src}->{$ver}->{'Build-Depends-Indep'}) {
		$bdi = $SCACHE->{$src}->{$ver}->{'Build-Depends-Indep'};
	}
	if (defined $SCACHE->{$src}->{$ver}->{'Build-Conflicts-Indep'}) {
		$bci = $SCACHE->{$src}->{$ver}->{'Build-Conflicts-Indep'};
	}
	return ($bd,$bc,$bdi,$bci);
}


sub S_getPkgField {
	my ($src,$ver,$field) = @_;
	showsrc2Cache($src);
	lvchomp($ver);
	my $ret;
	if (my $F = $SCACHE->{$src}->{$ver}->{$field}){
		foreach my $p (split /\s*,\s*/, $F){
				$ret->{$p} = 0;
		}
		return $ret;
	}
}

sub binSourceVField {
	my ($p,$v) = @_;
	#lvchomp($v);
	show2Cache($p);
	if ($PCACHE->{$p} && $PCACHE->{$p}->{$v} 
	&& $PCACHE->{$p}->{$v}->{'Source'} 
	&& ($PCACHE->{$p}->{$v}->{'Source'} =~ /\(([^\)]+)\)/)){
		return $1;
	}
	return $v;
}

sub binSourcePField {
	my ($p,$v) = @_;
	show2Cache($p);
	if ($PCACHE->{$p} && $PCACHE->{$p}->{$v} 
	&& $PCACHE->{$p}->{$v}->{'Source'} 
	&& ($PCACHE->{$p}->{$v}->{'Source'} =~ /^(\S+)/)){
		my $one = $1;
		showsrc2Cache($one);
		return $one;
	}
	return $p;
}

sub dbgCache {
	my $c = shift;
	foreach my $pp (sort keys %$c) {
		print "PP $pp\n";
		foreach my $vv (sort keys %{$c->{$pp}}){
			print "  VV $vv\n";
			foreach my $ff (sort keys %{$c->{$pp}->{$vv}}){
				print "     FF $ff\: ";
				print $c->{$pp}->{$vv}->{$ff},"\n";
			}
		}
	}
}

sub verifyDSC {
	return 1 unless (-r $_[0]);
	if (-x $DSCV && sys($DSCV,$_[0])==0){
		return 0;
	} else {
		return verifyWOSig($_[0]);
	}
}

sub verifyWOSig {
	my $fn = shift;
	my $fdir = dirname($fn);
	my $in = 0;
	my $fh = zOpen($fn);
	my @toCheck;
	while (<$fh>){
		if (/^Files:(.*)/i){
			mchomp(my $i = $1);
			$in = 1;
		} elsif ($in && 
		/^\s+([a-f\d]{32})\s+(\d+)\s+(?:\S+\s+\S+\s+)?(\S+)\s*$/) {
			my $ff = {
				'md5' => $1,
				'size' => $2,
				'file' => "$fdir/$3"
			};
			push @toCheck, $ff;
		}
	}
	close $fh;
	foreach my $ff (@toCheck) {
		return 1 unless (-r $ff->{'file'});
		return 1 unless ((-s $ff->{'file'}) == $ff->{'size'});
		return 1 unless (md5sum($ff->{'file'}) eq $ff->{'md5'});
		info(0,"   validated ", basename($ff->{'file'}),
		" (no signature verification)");
	}
	info(0,"Verified md5 sums and filesizes for files in: ", basename($fn),
	" (no signature verification)");

	return 0;
}

sub S_verSelect {
	# $p can be either a source or package name
	# $want: 'newest' or 'default' version according to apt-cache policy
	my ($p,$want)=@_;
	my $ret = $verDefault;
	mchomp($p);
	mchomp($want);
	#lvchomp($want);
	my $pvd = $verDefault;
	showsrc2Cache($p);
#	print "WWWWWWWWAAAAAAAAAANTTT $p $want\n" if ($p=~/gdk/);
	if ($want eq 'newest') { 
		$pvd = P_verNewest($p);
	#	print "NEWEST $p => $pvd\n";
	} elsif ($want eq 'default'){
		$pvd = P_verDefault($p);
	#	print "DEFAULT $p => $pvd\n";
	} else {
		$pvd = $want;
	}
#	print "PVDDDDDDDDDDDDDDDD $pvd\n" if ($p=~/gdk/);
	if ($pvd ne $verDefault) {
		#lvchomp($pvd);
		my $rv = binSourceVField($p,$pvd);
		my $sp = binSourcePField($p,$pvd);
		showsrc2Cache($sp);
		if (defined $SCACHE->{$sp}->{$rv}) {
			return ($sp,$rv) unless ($rv eq $verDefault);
		}
		lvchomp($rv);
		if (defined $SCACHE->{$sp}->{$rv}) {
			return ($sp,$rv) unless ($rv eq $verDefault);
		}
	}

	# the long way... (fallback method for finding source packages
	# that are no longer apt-getable)
	my $spkg = S_listBins($p);
	my @guess = ('',$verDefault,$verDefault);
	my @fb = ('',$verDefault);
	# O^3, but these are small loops
	lvchomp($want);
	foreach my $src (keys %$spkg) {
#	print "XXXXXX NEWEST1 ($p/$want) = $src\n" if ($p=~/libperl/);
		foreach my $ver (sort {	$VS->compare($b,$a)	} 
		keys %{$spkg->{$src}}){
			if ($VS->compare($ver,$fb[1]) > 0){
		#		print "XXXXXX NEWEST ($p/$want) = $src/$ver\n";
				@fb = ($src,$ver);
				return (@fb) if ($want eq 'newest');
			}
			foreach my $bin (sort keys %{$spkg->{$src}->{$ver}}){
$bin =~s/^__//;
my $cver;
#if ($want eq 'newest') { 
#	$cver = P_verNewest($bin);
#	$cver = binSourceVField($p,$cver);
#} elsif ($want eq 'default') {
if ($want eq 'default') {
	$cver = P_verDefault($bin);
	lvchomp($cver);
	#print "PVD: $bin ($cver)\n";
	$cver = binSourceVField($p,$cver);
} else {
	$cver = $want;
	$cver = binSourceVField($p,$cver);
}
if ($cver eq $ver) {
	return ($src,$ver) unless ($ver eq $verDefault);
} elsif ($cver eq $want) {
	# grab already downloaded versions (no longer in accesible archives)
	my $prefix = namePrefix($src);
	my $sdir = "$BUILDDIR/$prefix/$src";
	if (-d $sdir) {
		chdir $sdir;
		my($ep,$ups,$deb) = splitVer($cver);
		my $dsc = $src.'_'.$ups.$deb.'.dsc';
		if(-r $dsc && verifyDSC($dsc)==0){
			return ($src,$cver) unless ($cver eq $verDefault);
		}
	}
} 
if ($cver =~s/(\.aptfu\.\d+)$//) {
	if ($cver ne $ver){
		$cver .= $1;
	} else { 
		return ($src,$ver) unless ($ver eq $verDefault);
	}
}
if ( ($cver =~ /\Q$ver\E/)||($ver =~/\Q$cver\E/) ) {
	if ($cver eq $want){
#		wLog("Version approximation: $p (wanted: $want)/$src ($cver =~ $ver)");
		return ($src,$ver) unless ($ver eq $verDefault);
	}
	if ($VS->compare($cver,$guess[2])>0) {
		@guess = ($src,$ver,$cver)
	}
} elsif (
(my($cep,$cups,$cdeb)=splitVer($cver)) && (my($ep,$ups,$deb)=splitVer($ver))){
	if (($cep eq $ep) && ($cups eq $ups)){
		if ($VS->compare($cver,$guess[2])>0) {
			@guess = ($src,$ver,$cver);
		}
	} else {
		# not sure about this part, very much approximated versions...
		my @sp = split /[^A-Z0-9]/i, $ups;
		my @csp = split /[^A-Z0-9]/i, $cups;
		my $max = ($#sp < $#csp) ? $#sp : $#csp;
		if ($max > 0){
			my $i = 0;
			for ($i=0;$i<=$max;++$i){
				last if ($sp[$i] ne $csp[$i]);
			}
			if (($i/$max) > .5) {
				if ($VS->compare($cver,$guess[2])>0) {
					@guess = ($src,$ver,$cver);
				}
			}
		}
	}
} else {
	unless ($PCACHE->{$bin}->{'Cached'}){
		show2Cache($bin);
	}
	foreach my $pv (keys %{$PCACHE->{$bin}}){
		if ( ($pv eq $cver)  &&
		$PCACHE->{$bin}->{$pv}->{'Source'} &&
		($PCACHE->{$bin}->{$pv}->{'Source'}=~/(\S+)\s*\(([^\)]+)\)/)){
			return ($1,$2) unless ($2 eq $verDefault);
		}
	}
}
			}
		}
	}
	# fallback for looking at the dpkg status file	
	# sometimes the status file specifies a source version different
	# from the one installed
	cacheStatus();
	if ( (defined $PCACHE->{$p}->{'CurrentVer'}) && 
	( $PCACHE->{$p}->{'CurrentVer'} eq $ret ) &&
	( defined $PCACHE->{$p}->{'CurrentSrc'} ) &&
	( $PCACHE->{$p}->{'CurrentSrc'}=~/(\S+)\s*\(([^\)]+)\)/ ) ) {
		return($1,$2) unless ($2 eq $verDefault);
	}
	if (@guess) {
		my ($src,$ver,$cver)=@guess;
		unless ($ver eq $verDefault){
#wLog("Version (guess) approximation: $p (wanted:$want)/$src ($cver =~ $ver)");
			return ($src,$ver);
		}
	}
	if (($fb[1] eq $verDefault) && (isInstalled($p)!=0)){
		EXIT("BAD_VERSION ($p,$want): $fb[0]($fb[1])");
	}
#	unless ($fb[1] eq $verDefault){
#		print STDERR "newest chosen (@fb) instead of $want\n";
#	}
	return @fb;
}

# returns $vers, a hash ref with all source versions available
sub bin2srcV {
	my $p = shift;
	my $vers;
	my ($pkg,$foo) = S_verSelect($p,'newest');
	showsrc2Cache($pkg);
	foreach my $i (keys %{$SCACHE->{$pkg}}) {
		next if (defined $CacheField{$i});
		$vers->{$pkg}->{$i} = 1;
	}
	return ($pkg,$vers);
}

# memoize source info into a hash ref so we don't have to run the apt-cache
# binary repeatedly
sub showsrc2Cache {
	my $src = shift;
	$src=~s/\=\S+//g;
	$src=~ s/[^\w\d\+\.\-\s]+//g;
	return 0 if ($SCACHE->{$src}->{'Cached'});
	&initAptPkgCache;
	my $fh;
	open $fh,"$APTCACHE showsrc $src 2>/dev/null |" 
	|| EXIT("$!: showsrc $src");
	$SCACHE = ctrl2Href($fh,$SCACHE,$fS,'Package');
	$SCACHE->{$src}->{'Cached'} = 1;
}

# XXX note to self: gradually migrate to this one over the non-M version
sub showsrc2CacheM {
	my $ops = shift;
	my $ps;
	foreach my $p (keys %$ops){
		#$p =~ s/\=\S+//g;
		($p) = ($p =~ /^\s*([\w\d\+\.\-\s]+)/);
		$ps->{$p} = 1 unless (exists $SCACHE->{$p} 
		&& exists $SCACHE->{$p}->{'Cached'}
		&&( $SCACHE->{$p}->{'Cached'}==1));
	}
	my $arg = join(' ',keys %$ps);
	return unless ($arg);
	&initAptPkgCache;
	my $fh;
	#print "AC showsrc $arg\n" if ($arg=~/\S+\s+\S+/);
	open $fh,"$APTCACHE showsrc $arg 2>/dev/null|"||EXIT("$!: showsrc $arg");
	$SCACHE = ctrl2Href($fh,$SCACHE,$fS,'Package');
	# match the binaries, too
	foreach my $p (keys %$ps) { $SCACHE->{$p}->{'Cached'} = 1; }
}

# find the source packages that are either named $p or list
# $p as a binary package that is made from it.  Use the binary names
# to check for policy settings
sub S_listBins {
	my $p = shift;
	showsrc2Cache($p);
	my $spkg;
	my $bb;
	my $c = 0;
	foreach my $pkg (keys %$SCACHE) {
		next if (defined $CacheField{$pkg});
		foreach my $ver (keys %{$SCACHE->{$pkg}}) {
			next if (defined $CacheField{$ver});
			if (my $bin = $SCACHE->{$pkg}->{$ver}->{'Binary'}) {
				if (($pkg eq $p) ||
				($bin =~
				/(?:[^\w\d\+\-\.]|^)\Q$p\E(?:[^\w\d\+\-\.]|$)/)){
			#	($bin =~/$p/)){
					++$c;
					foreach my $e 
					(split(/\s*(?:,|\n)\s*/,$bin)){
						chomp $e;
						$spkg->{$pkg}->{$ver}->{$e} = 0;
						$bb->{$e} = 0 unless($bb->{$e});
						++$bb->{$e};
					}
				}
			}
		}
	}
	# put common binaries first...
	foreach my $e (keys %$bb) { 
		unless ($bb->{$e} == $c){
	#		print "DBG $e deleted ($bb->{$e} != $c)\n";
			delete $bb->{$e};
		}
	}
	foreach my $pp (keys %$spkg){
		foreach my $v ( keys %{$spkg->{$pp}}){
			foreach my $e (keys %$bb){
	#			print "COMMON IN ALL $c in $e ($p) [$pp($v)]\n";
				$spkg->{$pp}->{$v}->{"__$e"} = 0;
			}
		}
	}
	return $spkg;
}
1;

